/* ***** BEGIN LICENSE BLOCK *****
 * This file is part of Natron <https://natrongithub.github.io/>,
 * (C) 2018-2020 The Natron developers
 * (C) 2013-2018 INRIA and Alexandre Gauthier-Foichat
 *
 * Natron is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Natron is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Natron.  If not, see <http://www.gnu.org/licenses/gpl-2.0.html>
 * ***** END LICENSE BLOCK ***** */

#ifndef NATRON_ENGINE_MEMORYFILE_H
#define NATRON_ENGINE_MEMORYFILE_H

// ***** BEGIN PYTHON BLOCK *****
// from <https://docs.python.org/3/c-api/intro.html#include-files>:
// "Since Python may define some pre-processor definitions which affect the standard headers on some systems, you must include Python.h before any standard headers are included."
#include <Python.h>
// ***** END PYTHON BLOCK *****

#include "Global/Macros.h"

#include <string>

#include "Global/GlobalDefines.h"
#include "Global/Enums.h"

#include "Engine/EngineFwd.h"

NATRON_NAMESPACE_ENTER

struct MemoryFilePrivate;

/**
 * @brief A memory file wrapper that maps a file to the virtual memory of the process.
 * This is not MT-safe.
 **/
class MemoryFile
{
public:

    enum FileOpenModeEnum
    {
        eFileOpenModeCreate = 0,

        eFileOpenModeOpen,

        eFileOpenModeOpenOrCreate,

        eFileOpenModeOpenTruncate,

        eFileOpenModeOpenTruncateOrCreate
    };

    /**
     * @brief Creates an empty object. No file is created/opened and no mapping is effective yet.
     * You can then call open(...) to open a file and then resize(...) it as needed.
     **/
    MemoryFile();

    /**
     * @brief The destructor closes the mapping, removing the RAM portion but not the file.
     * The data are not guaranteed to be flushed to the disk.
     **/
    ~MemoryFile();

    /**
     * @brief Attempts to create the file if the file didn't exist already.
     * If the file exists already, this function will also map the file to memory and the data
     * can be read using the pointer returned by data(). Otherwise data() returns NULL and you
     * explicitly need to call resize(...) to create the file mapping.
     *
     * WARNING: Calling this function whilst the mapping is already opened has no effect
     * This function might throw an exception upon failure to open the file.
     **/
    void open(const std::string & filepath, FileOpenModeEnum open_mode);

    /**
     * @brief Returns a pointer to the beginning of the file,
     * if the file has been successfully opened, otherwise it returns 0.
     **/
    char* getData() const;

    /**
     * @brief Changes the number of bytes of the significant
     * part of the file. The resulting size can be retrieved
     * using the "size" function.
     *
     * WARNING: Any pointer held to data() previously to calling resize()
     * will be garbage after calling this.
     *
     * This function might throw an exception upon failure to map the file or truncate the file.
     **/
    void resize(size_t new_size, bool preserve);

    /**
     * @brief Close any mapping opened. Data will be not be flushed, make sure to call
     * flush first.
     **/
    void close();

    /**
     * @brief Closes the mapping if it was opened and re-open it.
     * If the mapping was not opened this function does nothing.
     * This is the same as flush + close + open
     **/
    void remap();

    /**
     * @brief Returns the size of the file in bytes.
     **/
    size_t size() const;

    enum FlushTypeEnum
    {
        eFlushTypeSync,
        eFlushTypeAsync,
        eFlushTypeInvalidate
    };
    /**
     * @brief Ensures that the backing file is in sync. with the data in memory
     * @param data If non null, only the portion starting at data and spanning size bytes
     * will be synced
     **/
    bool flush(FlushTypeEnum type, void* data, std::size_t size);

    /**
     * @brief Returns the filepath of the backing file.
     **/
    std::string path() const;

    /**
     * @brief Removes the backing file and closes the mapping to the virtual memory.
     * After that you could re-use the object calling the open(...) function again.
     **/
    void remove();

private:

    MemoryFilePrivate* _imp;
};

NATRON_NAMESPACE_EXIT

#endif /* defined(NATRON_ENGINE_MEMORYFILE_H_) */
